/* AVR RS485 Bootloader
 ****************************************************************************
 * Author:      Hendrik Hlscher                         					*
 * modified by:    															*
 * Revision:    0.9                              							*
 * Date:        11.05.21                  									*
 * Target:      Transceiver Rev.3.01 [ATmega8515]       			*
 * written by hendrik hoelscher, www.hoelscher-hi.de with ATMEL Studio 7                								*
 ****************************************************************************
 * Fuses:    OSC:  8.0 Mhz xtal	          								*
 *           WDT:  on                       								*
 ****************************************************************************
 *                   Software License Agreement                       		*
 *                                                                    		*												*
 * 					                              							*
 * THIS SOFTWARE IS PROVIDED IN AN "AS IS" CONDITION. NO WARRANTIES,   		*
 * WHETHER EXPRESS, IMPLIED OR STATUTORY, INCLUDING, BUT NOT LIMITED   		*
 * TO, IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A 	      		*
 * PARTICULAR PURPOSE APPLY TO THIS SOFTWARE. HENNE'S SITES SHALL NOT,   	*
 * IN ANY CIRCUMSTANCES, BE LIABLE FOR SPECIAL, INCIDENTAL OR 	      		*
 * CONSEQUENTIAL DAMAGES, FOR ANY REASON WHATSOEVER.	              		*
 ****************************************************************************
 */

#ifndef F_CPU
   #define  F_CPU	(8000000UL)
#endif

#include <avr/wdt.h>
#include <avr/pgmspace.h>
#include <avr/boot.h>
#include <util/delay.h>
#include <avr/interrupt.h>



#define PAGE_LIMIT		(120)					//this page and pages above are not rewritable (extra security for bootloader)
#define RX_ERR			(0xAA)
			uint8_t		Buffer[(SPM_PAGESIZE +1)]; //page data buffer
			

typedef union {
   uint16_t u16;
   struct {
          uint8_t u8l;
          uint8_t u8h;
          };
} uint16or8;



BOOTLOADER_SECTION int boot_main(void)
{
cli();

wdt_reset();
wdt_enable(WDTO_2S);							//enable watchdog (Timeout= 2s)
wdt_reset();

// **** INIT ****
//Ports
DDRA  = 0b11111111;								//low outputs
PORTA = 0b00000000;
DDRC  = 0b00000000;								//DIP switches with pullup
PORTC = 0b11111111;
DDRD  = 0b10010110;								//DMX, LED1, ZC
PORTD = 0b11111110;								//red LED off, transmission enabled
DDRE  = 0b00000001;								//bit9, option, LED0 (off)
PORTE = 0b00000110;

//USART
UBRRH  = 0;
UBRRL  = ((F_CPU/4000000UL)-1);					//250kbaud, 8n2
UCSRC  = (1<<URSEL)|(3<<UCSZ0)|(1<<USBS);


if ((PINC != 0xFF) || (PINE != 0b110))
	{
	asm ("ldi R31, 0x00");						//if one DIP switch is on -> exit bootloader
	asm ("ldi R30, 0x00");
	asm ("ijmp");
	}

uint8_t Data= SPM_PAGESIZE;

for(;;)
	{
	//**** send product ID or Response ****
	UCSRB  = (1<<TXEN);
	UDR= Data;
	loop_until_bit_is_set(UCSRA, TXC);				//wait until transmission complete
	_delay_us(90);
	UCSRB  = (1<<RXEN);								//enable receiver
	PORTD &= ~(1<<PD2);
	//**** wait till PC send 'P' ****
	Data= 0;
	while (Data != 'P') 						//wait till PC send 'P'
		{
		loop_until_bit_is_set(UCSRA, RXC);		//wait until reception complete				
		Data= UDR;
		}
	wdt_reset();
	
	//**** get Page ****
	uint8_t crc= 0;	
	uint8_t i;
	for (i=0; i< (SPM_PAGESIZE +2); i++)
		{
		loop_until_bit_is_set(UCSRA, RXC);		//wait until reception complete					
		if ((UCSRA &0b00011000) != 0)			//check for data overruns and framing errors
			{
			crc = RX_ERR;
			}
		Data= UDR;
		crc += Data;
		Buffer[i]= Data;						//fill buffer page				
		}
	if (crc == 0)
		{
		if (Buffer[0] < PAGE_LIMIT)
		   {
			//**** write Page ****
			if (Buffer[0] &(1<<0)) PORTE |=  (1<<PE0);	//toggle green LED to show activity
			else                   PORTE &= ~(1<<PE0);
			uint16_t Page_ex= Buffer[0] *SPM_PAGESIZE;	//get absolute base address
			boot_page_erase(Page_ex);				//erase page
			boot_spm_busy_wait();
			uint8_t k= 1;
			uint16or8 w;
			for (i=0; i<SPM_PAGESIZE; i+=2)
				{
				w.u8l= Buffer[k++];
				w.u8h= Buffer[k++];
				boot_page_fill(Page_ex +i, w.u16);	//load new page
				}
			boot_page_write(Page_ex);				//store
			boot_spm_busy_wait();
			boot_rww_enable();						//re-enable normal flash operation
			//**** check page ****
			Data= 0;								//return OK
			for (i= 0; i<SPM_PAGESIZE; i++)
				{
				if (Buffer[1+ i] != pgm_read_byte(Page_ex +i))
					{
					Data= 3;						//return ERR
					}
				}
		   }
		else Data= 1;								//return write in boot sector ERROR
		}
	else Data= 2;									//return check sum ERROR
	
	//**** prepare response ****
	PORTD |= (1<<PD2);								//enable transmitter
	if (Data == 0) PORTD |=  (1<<PD7);			
	else           PORTD &= ~(1<<PD7);				//enable red LED on error
	_delay_ms(2);
	}
}
