// Debug code for Receiver Rev.3.01
// release date: 13.09.06
// target: ATmega8515 @8MHz
// Author: Hendrik Hlscher
// URL: www.hoelscher-hi.de
// (c)H.Hlscher, all rights reserved

#include <avr/io.h>
#include <stdint.h>
#include <avr/interrupt.h>
#include <avr/wdt.h>

#define  F_CPU	(8000000UL)

//global vars
volatile	uint8_t  Flags;
volatile    uint8_t  FlagsBuf;
            uint8_t  DmxState;

enum {FUSE_OK, SIGNAL_OK, DMX_OK, ZC_OK};
enum {IDLE, BREAK, STARTB};				//DMX states


void init(void)
{
	wdt_reset();
	wdt_enable(WDTO_120MS);			//enable watchdog
	wdt_reset();
	
	DDRA  = 0xFF;					//low outputs
	PORTA = 0;

	DDRC  = 0;						//DIP switches with pullup
	PORTC = 0xFF;

	DDRD  = 0b10010100;				//DMX, LED1, ZC
	PORTD = 0b01111000;				//red LED on

	DDRE  = 0b00000001;				//bit9, option, LED0 (off)
	PORTE = 0xFF;

	TCCR1A = 0;
	TCCR1B = (1<<CS10)|(1<<CS12)|(1<<WGM12); //set T1 to clk/1024, CTC mode
	OCR1A  = 651;					//set IRQ to 12Hz
	TIMSK  = (1<<OCIE1A);

	GICR   = (1<<INT1);				//enable ext int1 for zero crossing detection
	MCUCR  = (1<<ISC11);			//interrupt on falling edge

	UBRRH  = 0;
	UBRRL  = 1;						//250kbaud, 8N2
	UCSRC  = (1<<URSEL)|(3<<UCSZ0)|(1<<USBS);
	UCSRB  = (1<<RXEN)|(1<<RXCIE);

	DmxState= IDLE;
	Flags =0;
}


int main(void)
{
	uint16_t adr;
	uint8_t  led;									//led= 0: green LED off
	cli();
	init();
	sei();
	while (1)
	   {
		adr= 0xFF -PINC;							//get address from DIP switches
		if (!(PINE &(1<<PE1))) adr += 0x100;
		if (!(PINE &(1<<PE2))) adr += 0x200;
		
		led= 0;
		if (adr == 0x000)
		   {
		   led= 1;
		   }
		else if (adr == 0x3FF)
		   {
		   led= 1;
		   }
		else if (adr == 0b0001)
		   {
		   if (FlagsBuf &(1<<FUSE_OK)) led= 1;  
		   }
		else if (adr == 0b0010)
		   {
		   if (FlagsBuf &(1<<SIGNAL_OK)) led= 1;
		   }
		else if (adr == 0b0100)
		   {
		   if (FlagsBuf &(1<<DMX_OK)) led= 1;
		   }
		else if (adr == 0b1000)
		   {
		   if (FlagsBuf &(1<<ZC_OK)) led= 1;
		   }
		   
		if (led == 0) PORTE |=  (1<<PE0);		//disable green LED
		else          PORTE &= ~(1<<PE0);		//enable green LED
	   }
};


// indicator routine (called with 12Hz)
ISR (TIMER1_COMPA_vect)
{	
	wdt_reset();
	PORTD ^= (1<<PD7);				//blink red LED with 12Hz
    
	FlagsBuf= Flags;				//double buffer flags to prevent flicker
	Flags   = (1<<FUSE_OK);			//mark ISR was triggered (and clear all other flags)
}


//ext irq caused by zc
ISR (INT1_vect)
{
	Flags |= (1<<ZC_OK);
}


//DMX reception
ISR (USART_RX_vect)
{
	uint8_t state= UCSRA;				//get state
	uint8_t DmxByte= UDR;				//get data

	Flags |= (1<<SIGNAL_OK);

	if (state &(1<<FE))					//check for break
	   {
		UCSRA &= ~(1<<FE);				//reset flag
		DmxState= BREAK;				//got break
		return;
	   }
	
	if (DmxState == BREAK)
	   {
		if (DmxByte == 0) DmxState= STARTB; //got correct start code
		else              DmxState= IDLE;
	   }

	else if (DmxState == STARTB)
	   {
		Flags |=(1<<DMX_OK);			//got 1st byte (not used)
		DmxState= IDLE;					//wait for next break
	   }
}

